#include "helpers.h"
#include <math.h>
#include <stdio.h>

// Convert image to grayscale
void grayscale(int height, int width, RGBTRIPLE image[height][width])
{
    double red;
    double green;
    double blue;
    float average;

    for (int n = 0; n < height; n++)
    {
        for (int m = 0; m < width; m++)
        {
            red =  image[n][m].rgbtRed;
            green =  image[n][m].rgbtGreen;
            blue =  image[n][m].rgbtBlue;

            // Calculate average from the three integers
            average = ((red + green + blue) / 3);

            image[n][m].rgbtRed = round(average);
            image[n][m].rgbtGreen = round(average);
            image[n][m].rgbtBlue = round(average);
        }
    }
    return;
}



// Reflect image horizontally
void reflect(int height, int width, RGBTRIPLE image[height][width])
{
    RGBTRIPLE temp[height][width];

    // Make copy of image
    for (int n = 0; n < height; n++)
    {
        for (int m = 0; m < width; m++)
        {
            temp[n][m] = image[n][m];
        }
    }

    // Iterate through pixel positions and replace the width value with width (-1 to start at end) - m
    for (int n = 0; n < height; n++)
    {
        for (int m = 0; m < width; m++)
        {
            image[n][m] = temp[n][(width - 1) - m];
        }
    }
    return;
}



// Blur image
void blur(int height, int width, RGBTRIPLE image[height][width])
{
    RGBTRIPLE temp[height][width];
    double red;
    double green;
    double blue;
    int count;
    float average_red;
    float average_green;
    float average_blue;

    // make a temporary copy of image
    for (int n = 0; n < height; n++)
    {
        for (int m = 0; m < width; m++)
        {
            temp[n][m] = image[n][m];
        }
    }

    // loop through images pixels
    for (int a = 0; a < height; a++)
    {
        for (int b = 0; b < width; b++)
        {
            // Set colors back to 0
            red = 0;
            green = 0;
            blue = 0;

            // Set count back to 0
            count = 0;

            // if normal there is room for a 9-pixel box then implement it
            if (a > 0 && a < height - 1 && b > 0 && b < width - 1)
            {
                for (int q = 0; q < 3; q++)
                {
                    for (int r = 0; r < 3; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on left border
            if (a > 0 && a < height - 1 && b == 0)
            {
                for (int q = 0; q < 3; q++)
                {
                    for (int r = 1; r < 3; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on right border
            if (a > 0 && a < height - 1 && b == width - 1)
            {
                for (int q = 0; q < 3; q++)
                {
                    for (int r = 0; r < 2; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on top border
            if (a == 0 && b > 0 && b < width - 1)
            {
                for (int q = 1; q < 3; q++)
                {
                    for (int r = 0; r < 3; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on bottom border
            if (a == height - 1 && b > 0 && b < width - 1)
            {
                for (int q = 0; q < 2; q++)
                {
                    for (int r = 0; r < 3; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on top-left corner
            if (a == 0 && b == 0)
            {
                for (int q = 1; q < 3; q++)
                {
                    for (int r = 1; r < 3; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on top-right corner
            if (a == 0 && b == width - 1)
            {
                for (int q = 1; q < 3; q++)
                {
                    for (int r = 0; r < 2; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on bottom-right corner
            if (a == height - 1 && b == width - 1)
            {
                for (int q = 0; q < 2; q++)
                {
                    for (int r = 0; r < 2; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // if pixel is on bottom-left corner
            if (a == height - 1 && b == 0)
            {
                for (int q = 0; q < 2; q++)
                {
                    for (int r = 1; r < 3; r++)
                    {
                        red += temp[(a - 1) + q][(b - 1) + r].rgbtRed;
                        green += temp[(a - 1) + q][(b - 1) + r].rgbtGreen;
                        blue += temp[(a - 1) + q][(b - 1) + r].rgbtBlue;
                        count++;
                    }
                }
            }

            // Calculate averages
            average_red = red / count;
            average_green = green / count;
            average_blue = blue / count;

            // Assign new averaged values to original pixel
            image[a][b].rgbtRed = round(average_red);
            image[a][b].rgbtGreen = round(average_green);
            image[a][b].rgbtBlue = round(average_blue);
        }
    }
    return;
}



// Detect edges
void edges(int height, int width, RGBTRIPLE image[height][width])
{
    RGBTRIPLE temp[height][width];
    int gx_red;
    int gx_green;
    int gx_blue;
    int gy_red;
    int gy_green;
    int gy_blue;
    double new_red;
    double new_green;
    double new_blue;

    // Gx Matrix - detects vertical edges
    int gx[3][3];
    gx[0][0] = -1;
    gx[0][1] = 0;
    gx[0][2] = 1;
    gx[1][0] = -2;
    gx[1][1] = 0;
    gx[1][2] = 2;
    gx[2][0] = -1;
    gx[2][1] = 0;
    gx[2][2] = 1;

    // Gy Matrix - detects horizontal edges
    int gy[3][3];
    gy[0][0] = -1;
    gy[0][1] = -2;
    gy[0][2] = -1;
    gy[1][0] = 0;
    gy[1][1] = 0;
    gy[1][2] = 0;
    gy[2][0] = 1;
    gy[2][1] = 2;
    gy[2][2] = 1;


    // make a temporary copy of image
    for (int n = 0; n < height; n++)
    {
        for (int m = 0; m < width; m++)
        {
            temp[n][m] = image[n][m];
        }
    }

    // loop through images pixels
    for (int a = 0; a < height; a++)
    {
        for (int b = 0; b < width; b++)
        {
            // set all values to 0
            gx_red = 0;
            gx_green = 0;
            gx_blue = 0;
            gy_red = 0;
            gy_green = 0;
            gy_blue = 0;

            // if normal there is room for a 9-pixel box then implement it normally
            if (a > 0 && a < height - 1 && b > 0 && b < width - 1)
            {
                for (int q = 0; q < 3; q++)
                {
                    for (int r = 0; r < 3; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }
            }

            // if pixel is on left border
            if (a > 0 && a < height - 1 && b == 0)
            {
                for (int q = 0; q < 3; q++)
                {
                    for (int r = 1; r < 3; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }

            }


            // if pixel is on right border
            if (a > 0 && a < height - 1 && b == width - 1)
            {
                for (int q = 0; q < 3; q++)
                {
                    for (int r = 0; r < 2; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }
            }


            // if pixel is on top border
            if (a == 0 && b > 0 && b < width - 1)
            {
                for (int q = 1; q < 3; q++)
                {
                    for (int r = 0; r < 3; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }

            }


            // if pixel is on bottom border
            if (a == height - 1 && b > 0 && b < width - 1)
            {
                for (int q = 0; q < 2; q++)
                {
                    for (int r = 0; r < 3; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }
            }

            // if pixel is on top-left corner
            if (a == 0 && b == 0)
            {
                for (int q = 1; q < 3; q++)
                {
                    for (int r = 1; r < 3; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }
            }

            // if pixel is on top-right corner
            if (a == 0 && b == width - 1)
            {
                for (int q = 1; q < 3; q++)
                {
                    for (int r = 0; r < 2; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }
            }

            // if pixel is on bottom-right corner
            if (a == height - 1 && b == width - 1)
            {
                for (int q = 0; q < 2; q++)
                {
                    for (int r = 0; r < 2; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }
            }

            // if pixel is on bottom-left corner
            if (a == height - 1 && b == 0)
            {
                for (int q = 0; q < 2; q++)
                {
                    for (int r = 1; r < 3; r++)
                    {
                        // Multiply 9-pixel matrix around pixel of interest with the Gx and Gy matrices
                        gx_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gx[q][r]);
                        gx_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gx[q][r]);
                        gx_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gx[q][r]);

                        gy_red += (temp[(a - 1) + q][(b - 1) + r].rgbtRed * gy[q][r]);
                        gy_green += (temp[(a - 1) + q][(b - 1) + r].rgbtGreen * gy[q][r]);
                        gy_blue += (temp[(a - 1) + q][(b - 1) + r].rgbtBlue * gy[q][r]);
                    }
                }
            }

            // Calculate final values for Red Green and Blue: value = square root of GX^2 + GY^2 - ensures that values are all positive
            // All performed in double and rounded at the end (see bottom)
            // If the final value is above 255, then set it to 255
            new_red = sqrt(pow((double)gx_red, 2.0) + pow((double)gy_red, 2.0));
            if (new_red > 255)
            {
                new_red = 255;
            }

            new_green = sqrt(pow((double)gx_green, 2.0) + pow((double)gy_green, 2.0));
            if (new_green > 255)
            {
                new_green = 255;
            }

            new_blue = sqrt(pow((double)gx_blue, 2.0) + pow((double)gy_blue, 2.0));
            if (new_blue > 255)
            {
                new_blue = 255;
            }

            image[a][b].rgbtRed = round(new_red);
            image[a][b].rgbtGreen = round(new_green);
            image[a][b].rgbtBlue = round(new_blue);
        }
    }
    return;
}
